"""
mc_renderer: Monte Carlo basis generation for muography.

Usage
-----
    from ncuhep.muography.mc_renderer import MonteCarloRenderer
"""

from .monte_carlo import MonteCarloRenderer

__all__ = ["MonteCarloRenderer"]
